/* Copyright (c) 1997-2002 Miller Puckette and others.
* For information on usage and redistribution, and for a DISCLAIMER OF ALL
* WARRANTIES, see the file, "LICENSE.txt," in this distribution.  */

#ifndef __TEXT_H__
#define __TEXT_H__

char *text_symbolname(t_symbol *s, char *nulname);
char *text_ordinal(int n);

#endif
